// -*- c++ -*- *******************************************************
// Copyright (C) 2003 Sandia Corporation
// Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that this Notice and any statement
// of authorship are reproduced on all copies.
//
// $Id: ocaProgramIFFT1c.cxx,v 1.6 2008-01-25 20:31:25 kmorel Exp $

#include "ocaProgramIFFT1c.h"

#include "ocaProgramFFT1c.h"
#include "ocaLookUpBuffer.h"
#include "ocaDrawableBuffer.h"
#include "ocaFactory.h"
#include "ocaProgramPassThrough.h"
#include "ocaCgError.h"

#define GET_VERT_PROGRAM_PARAMETER(paramname, parameter)		\
    parameter = cgGetNamedParameter(this->vertProgram, paramname);	\
    if (parameter == NULL) {						\
	ocaRaiseError("Could not get pointer for parameter " #paramname); \
    }
#define GET_FRAG_PROGRAM_PARAMETER(paramname, parameter)		\
    parameter = cgGetNamedParameter(this->fragProgram, paramname);	\
    if (parameter == NULL) {						\
	ocaRaiseError("Could not get pointer for parameter " #paramname); \
    }

// Sometimes the Cg compile optimizes away variables that are not used.
// If you try to set them, you may get an error.  Ignore these errors.
#define IGNORE_INVALID_PARAMETER(command)                               \
    try {                                                               \
        command;                                                        \
    } catch (ocaCgError &error) {                                       \
        if (error.getError() != CG_INVALID_PARAMETER_ERROR) {           \
            throw error;                                                \
        }                                                               \
    }


//-----------------------------------------------------------------------------
// Auxiliary classes for tangling frequencies.

void ocaProgramIFFT1cTanglers::load_program(const char **compilerArgs)
{
    this->loadProgram("fft1c_tangle.cg", "FFT1cTangleVert",
		      "FFT1cTangleFrag", compilerArgs);
    GET_VERT_PROGRAM_PARAMETER("input.realFIndex", this->realFIndex);
    GET_VERT_PROGRAM_PARAMETER("input.imagFIndex", this->imagFIndex);
    GET_VERT_PROGRAM_PARAMETER("input.realGIndex", this->realGIndex);
    GET_VERT_PROGRAM_PARAMETER("input.imagGIndex", this->imagGIndex);
    GET_VERT_PROGRAM_PARAMETER("ArraySize", this->ArraySize);
    GET_FRAG_PROGRAM_PARAMETER("CompressedFrequencies",
			       this->CompressedFrequencies);
}
void ocaProgramIFFT1cTanglers::execute(ocaLookUpBuffer::pointer source,
				       const float *realFRegion,
				       const float *realGRegion,
				       const float *imagFRegion,
				       const float *imagGRegion)
{
    const int *arraySize = source->getSize();

    this->bind();
    cgGLSetParameter2f(this->ArraySize,
		       (float)arraySize[0], (float)arraySize[1]);
    cgGLSetTextureParameter(this->CompressedFrequencies,
			    source->getTextureId());
    cgGLEnableTextureParameter(this->CompressedFrequencies);

    IGNORE_INVALID_PARAMETER(cgGLSetParameterPointer(this->realFIndex, 2,
                                                     GL_FLOAT, 0,
                                                     (void *)realFRegion));
    IGNORE_INVALID_PARAMETER(cgGLSetParameterPointer(this->realGIndex, 2,
                                                     GL_FLOAT, 0,
                                                     (void *)realGRegion));
    IGNORE_INVALID_PARAMETER(cgGLSetParameterPointer(this->imagFIndex, 2,
                                                     GL_FLOAT, 0,
                                                     (void *)imagFRegion));
    IGNORE_INVALID_PARAMETER(cgGLSetParameterPointer(this->imagGIndex, 2,
                                                     GL_FLOAT, 0,
                                                     (void *)imagGRegion));

    IGNORE_INVALID_PARAMETER(cgGLEnableClientState(this->realFIndex));
    IGNORE_INVALID_PARAMETER(cgGLEnableClientState(this->imagFIndex));
    IGNORE_INVALID_PARAMETER(cgGLEnableClientState(this->realGIndex));
    IGNORE_INVALID_PARAMETER(cgGLEnableClientState(this->imagGIndex));

    glDrawArrays(GL_QUADS, 0, 4);

    IGNORE_INVALID_PARAMETER(cgGLDisableClientState(this->realFIndex));
    IGNORE_INVALID_PARAMETER(cgGLDisableClientState(this->imagFIndex));
    IGNORE_INVALID_PARAMETER(cgGLDisableClientState(this->realGIndex));
    IGNORE_INVALID_PARAMETER(cgGLDisableClientState(this->imagGIndex));
    cgGLDisableTextureParameter(this->CompressedFrequencies);

    this->unbind();
}

ocaProgramIFFT1cTangleRealLower::ocaProgramIFFT1cTangleRealLower() {
    this->load_program(this->compilerArgs);
}
ocaProgramIFFT1cTangleRealLower::~ocaProgramIFFT1cTangleRealLower() {}
const char *ocaProgramIFFT1cTangleRealLower::compilerArgs[] = {
    "-DCALCULATE_REAL", "-DCALCULATE_LOWER", NULL
};

ocaProgramIFFT1cTangleRealUpper::ocaProgramIFFT1cTangleRealUpper() {
    this->load_program(this->compilerArgs);
}
ocaProgramIFFT1cTangleRealUpper::~ocaProgramIFFT1cTangleRealUpper() {}
const char *ocaProgramIFFT1cTangleRealUpper::compilerArgs[] = {
    "-DCALCULATE_REAL", "-DCALCULATE_UPPER", NULL
};

ocaProgramIFFT1cTangleImagLower::ocaProgramIFFT1cTangleImagLower() {
    this->load_program(this->compilerArgs);
}
ocaProgramIFFT1cTangleImagLower::~ocaProgramIFFT1cTangleImagLower() {}
const char *ocaProgramIFFT1cTangleImagLower::compilerArgs[] = {
    "-DCALCULATE_IMAGINARY", "-DCALCULATE_LOWER", NULL
};

ocaProgramIFFT1cTangleImagUpper::ocaProgramIFFT1cTangleImagUpper() {
    this->load_program(this->compilerArgs);
}
ocaProgramIFFT1cTangleImagUpper::~ocaProgramIFFT1cTangleImagUpper() {}
const char *ocaProgramIFFT1cTangleImagUpper::compilerArgs[] = {
    "-DCALCULATE_IMAGINARY", "-DCALCULATE_UPPER", NULL
};


//-----------------------------------------------------------------------------
// Functions for main ocaProgramIFFT1c program.

ocaProgramIFFT1c::ocaProgramIFFT1c() {
}
ocaProgramIFFT1c::~ocaProgramIFFT1c() {
}

void ocaProgramIFFT1c::execute(ocaLookUpBuffer::pointer frequencies,
			       ocaDrawableBuffer::pointer samples)
{
    const int *size = samples->getSize();
    int vectorSize = samples->getVectorSize();

    ocaFactory::pointer factory = ocaFactory::getSingleton();
    ocaDrawableBuffer::pointer temporary
	= factory->makeDrawableBuffer(size[0], size[1], vectorSize);

    this->execute(frequencies, samples, temporary);
}

void ocaProgramIFFT1c::execute(ocaLookUpBuffer::pointer frequencies,
			       ocaDrawableBuffer::pointer samples,
			       ocaDrawableBuffer::pointer temporary)
{
    int size = frequencies->getSize()[0];
    int logsize = 0;
    int i = size;
    while (1) {
	if (i < 2) break;
	if ((i%2) == 1) ocaRaiseError("IFFT input must be a power of 2");
	i >>= 1;
	logsize++;
    }

    if ((logsize%2) == 0) {
	this->do_IFFT(frequencies, samples, temporary);
    } else {
	this->do_IFFT(frequencies, temporary, samples);
    }
}

ocaDrawableBuffer::pointer ocaProgramIFFT1c
    ::execute(ocaDrawableBuffer::pointer frequencies)
{
    const int *size = frequencies->getSize();
    int vectorSize = frequencies->getVectorSize();

    ocaFactory::pointer factory = ocaFactory::getSingleton();
    ocaDrawableBuffer::pointer temporary
	= factory->makeDrawableBuffer(size[0], size[1], vectorSize);

    return this->execute(frequencies, temporary);
}

ocaDrawableBuffer::pointer ocaProgramIFFT1c
    ::execute(ocaDrawableBuffer::pointer frequencies,
	      ocaDrawableBuffer::pointer auxiliary)
{
    ocaDrawableBuffer::pointer samples, other;
    this->execute(frequencies, auxiliary, samples, other);
    return samples;
}

void ocaProgramIFFT1c::execute(ocaDrawableBuffer::pointer frequencies,
			       ocaDrawableBuffer::pointer auxiliary,
			       ocaDrawableBuffer::pointer &samples,
			       ocaDrawableBuffer::pointer &other)
{
    ocaDrawableBuffer::pointer target = auxiliary;
    ocaDrawableBuffer::pointer feedback = frequencies;

    this->do_IFFT(feedback->getSharedLookUpBuffer(), target, feedback);

    samples = feedback;
    other = target;
}

void ocaProgramIFFT1c::do_IFFT(ocaLookUpBuffer::pointer source,
			       ocaDrawableBuffer::pointer &target,
			       ocaDrawableBuffer::pointer &feedback)
{
    const int *arraySize = source->getSize();

    int size = arraySize[0];
    int logsize = 0;
    int i = size;
    while (1) {
	if (i < 2) break;
	if ((i%2) == 1) ocaRaiseError("IFFT input must be a power of 2");
	i >>= 1;
	logsize++;
    }

  // Special case of 1 sample inputs.
    if (logsize == 0) {
	target->copy(ocaCast<ocaBuffer>(source));
	ocaDrawableBuffer::pointer swap = target;
	target = feedback;
	feedback = swap;
	return;
    }

    if (arraySize[1] < 2) {
	ocaRaiseError("Compressed IFFT requires at least 2 rows.");
    }
    if (arraySize[1]%2 != 0) {
	ocaRaiseError("Compressed IFFT requires an even amount of rows.");
    }

    const int *as_t = target->getSize();
    const int *as_f = feedback->getSize();
    if (   (arraySize[0] != as_t[0]) || (arraySize[1] != as_t[1])
	|| (arraySize[0] != as_f[0]) || (arraySize[1] != as_f[1]) ) {
	ocaRaiseError("Inconsistent buffer sizes.");
    }

  // Do the tangling.
    target->makeCurrent();

  // Tangle conjugates.
    ocaProgramIFFT1cTangleRealLower::pointer tangleRealLower
	= ocaProgramIFFT1cTangleRealLower::getSingleton();
    ocaProgramIFFT1cTangleRealUpper::pointer tangleRealUpper
	= ocaProgramIFFT1cTangleRealUpper::getSingleton();
    ocaProgramIFFT1cTangleImagLower::pointer tangleImagLower
	= ocaProgramIFFT1cTangleImagLower::getSingleton();
    ocaProgramIFFT1cTangleImagUpper::pointer tangleImagUpper
	= ocaProgramIFFT1cTangleImagUpper::getSingleton();

    float realFRegion[] = {
	1, 				0,
	(float)(arraySize[0]/2),	0,
	(float)(arraySize[0]/2),	(float)(arraySize[1]/2),
	1,				(float)(arraySize[1]/2)
    };
    float imagFRegion[] = {
	(float)(arraySize[0]),		0,
	(float)(arraySize[0]/2+1),	0,
	(float)(arraySize[0]/2+1),	(float)(arraySize[1]/2),
	(float)(arraySize[0]),		(float)(arraySize[1]/2)
    };
    float realGRegion[] = {
	1,				(float)(arraySize[1]/2),
	(float)(arraySize[0]/2),	(float)(arraySize[1]/2),
	(float)(arraySize[0]/2),	(float)(arraySize[1]),
	1,				(float)(arraySize[1])
    };
    float imagGRegion[] = {
	(float)(arraySize[0]),		(float)(arraySize[1]/2),
	(float)(arraySize[0]/2+1),	(float)(arraySize[1]/2),
	(float)(arraySize[0]/2+1),	(float)(arraySize[1]),
	(float)(arraySize[0]),		(float)(arraySize[1])
    };

    tangleRealLower->execute(source, realFRegion, realGRegion,
			     imagFRegion, imagGRegion);
    tangleRealUpper->execute(source, realFRegion, realGRegion,
			     imagFRegion, imagGRegion);
    tangleImagLower->execute(source, realFRegion, realGRegion,
			     imagFRegion, imagGRegion);
    tangleImagUpper->execute(source, realFRegion, realGRegion,
			     imagFRegion, imagGRegion);

  // Pass through stuff that doesn't need changing.
    const static float PassThroughLines[] = {
	-1, -1,
	-1, 1,
	0, -1,
	0, 1
    };

    ocaProgramPassThrough::pointer passthrough
	= ocaProgramPassThrough::getSingleton();
    passthrough->bind();
    cgGLSetTextureParameter(passthrough->getSource(), source->getTextureId());
    cgGLEnableTextureParameter(passthrough->getSource());
    cgGLSetParameterPointer(passthrough->getPosition(), 2, GL_FLOAT, 0,
			    (void *)PassThroughLines);
    cgGLEnableClientState(passthrough->getPosition());
    glDrawArrays(GL_LINES, 0, 4);
    cgGLDisableClientState(passthrough->getPosition());
    cgGLDisableTextureParameter(passthrough->getSource());
    passthrough->unbind();

  // In case source came from feedback.
    feedback->releaseLookUpBuffer();

  // Swap target & feedback.
    ocaDrawableBuffer::pointer swap = target;
    target = feedback;
    feedback = swap;

  // Do the actuall FFT algorithm.
    ocaProgramFFT1cCore::getSingleton()
	->execute(feedback->getSharedLookUpBuffer(), target, feedback, false);
}
