// -*- c++ -*- *******************************************************
// Copyright (C) 2003 Sandia Corporation
// Under the terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that this Notice and any statement
// of authorship are reproduced on all copies.
//
// $Id: ocaProgramMatrixMultiply.cxx,v 1.8 2003-06-30 17:56:02 kmorel Exp $

#include "ocaProgramMatrixMultiply.h"

#include "ocaLookUpBuffer.h"
#include "ocaDrawableBuffer.h"
#include "ocaZeroBuffer.h"
#include "ocaFactory.h"

#include <iostream>

#define GET_PROGRAM_PARAMETER(program, paramname, parameter)		\
    parameter = cgGetNamedParameter(program, paramname);		\
    if (parameter == NULL) {						\
	ocaRaiseError("Could not get pointer for parameter " #paramname); \
    }

ocaProgramMatrixMultiply::ocaProgramMatrixMultiply()
{
    this->loadProgram("matrixmultiply.cg",
		      "MatrixMultiplyVert", "MatrixMultiplyFrag");

    GET_PROGRAM_PARAMETER(this->vertProgram, "input.position", this->position);
    GET_PROGRAM_PARAMETER(this->vertProgram, "input.index", this->index);

    GET_PROGRAM_PARAMETER(this->fragProgram, "A", this->matrixA);
    GET_PROGRAM_PARAMETER(this->fragProgram, "B", this->matrixB);
    GET_PROGRAM_PARAMETER(this->fragProgram, "ACCUM", this->matrixACCUM);
}

ocaProgramMatrixMultiply::~ocaProgramMatrixMultiply()
{
}

void ocaProgramMatrixMultiply::execute(ocaLookUpBuffer::pointer matrixA,
				       ocaLookUpBuffer::pointer matrixB,
				       ocaDrawableBuffer::pointer destination,
				       bool ARowIsT, bool BRowIsT,
				       bool DestRowIsY)
{
    const int *size = destination->getSize();
    int vs = destination->getVectorSize();

    ocaDrawableBuffer::pointer temporary(NULL);
    if (   ( ARowIsT && (matrixA->getSize()[0] != 1))
	|| (!ARowIsT && (matrixA->getSize()[1] != 1)) ) {
	temporary
	    = ocaFactory::getSingleton()->makeDrawableBuffer(size[0], size[1],
							     vs);
    }
    this->execute(matrixA, matrixB, destination, temporary, ARowIsT, BRowIsT,
		  DestRowIsY);
}

static GLfloat fullScreenSquare[] = {
    -1.0, -1.0,
    1.0, -1.0,
    1.0, 1.0,
    -1.0, 1.0
};

void ocaProgramMatrixMultiply::execute(ocaLookUpBuffer::pointer matrixA,
				       ocaLookUpBuffer::pointer matrixB,
				       ocaDrawableBuffer::pointer destination,
				       ocaDrawableBuffer::pointer temporary,
				       bool ARowIsT, bool BRowIsT,
				       bool DestRowIsY)
{
    GLfloat matrixIndices[4*4];

    if (!temporary.isNull()) {
	const int *size1, *size2;
	size1 = destination->getSize();
	size2 = temporary->getSize();

	if ((size1[0] != size2[0]) || (size1[1] != size2[1])) {
	    ocaRaiseError("Temporary drawable buffer does not have same "
			  "dimensions as destination.");
	}
    }

  // Get matrix dimensions and make sure they are consistent.
    int rows, cols;
    int intermediate;
    int size[2];
    if (DestRowIsY) {
	destination->ocaBuffer::getSize(cols, rows);
    } else {
	destination->ocaBuffer::getSize(rows, cols);
    }

    if (ARowIsT) {
	matrixA->ocaBuffer::getSize(intermediate, size[0]);
    } else {
	matrixA->ocaBuffer::getSize(size[0], intermediate);
    }
    if (size[0] != rows) {
	ocaRaiseError("Matrix A does not agree with output dimensions.");
    }

    if (BRowIsT) {
	matrixB->ocaBuffer::getSize(size[1], size[0]);
    } else {
	matrixB->ocaBuffer::getSize(size[0], size[1]);
    }
    if (size[1] != cols) {
	ocaRaiseError("Matrix B does not agree with output dimensions.");
    }
    if (size[0] != intermediate) {
	ocaRaiseError("Matrix A and matrix B do not have agreeable dimensions");
    }

  // Establish input textures in both drawables.
    destination->makeCurrent();
    this->bind();
    cgGLSetTextureParameter(this->matrixA, matrixA->getTextureId());
    cgGLEnableTextureParameter(this->matrixA);
    cgGLSetTextureParameter(this->matrixB, matrixB->getTextureId());
    cgGLEnableTextureParameter(this->matrixB);
    cgGLSetParameterPointer(this->position, 2, GL_FLOAT, 0, fullScreenSquare);
    cgGLEnableClientState(this->position);
    cgGLSetParameterPointer(this->index, 4, GL_FLOAT, 0, matrixIndices);
    cgGLEnableClientState(this->index);


    if (!temporary.isNull()) {
	temporary->makeCurrent();
	this->bind();
	cgGLSetTextureParameter(this->matrixA, matrixA->getTextureId());
	cgGLEnableTextureParameter(this->matrixA);
	cgGLSetTextureParameter(this->matrixB, matrixB->getTextureId());
	cgGLEnableTextureParameter(this->matrixB);
	cgGLSetParameterPointer(this->position, 2, GL_FLOAT, 0,
				fullScreenSquare);
	cgGLEnableClientState(this->position);
	cgGLSetParameterPointer(this->index, 4, GL_FLOAT, 0, matrixIndices);
	cgGLEnableClientState(this->index);

    }

  // Set up output matrices so that the final output is contained in
  // destination.
    ocaDrawableBuffer::pointer feedback;
    ocaDrawableBuffer::pointer target;
    if (intermediate%2 == 1) {
	target = destination;
	feedback = temporary;
    } else {
	target = temporary;
	feedback = destination;
    }

    for (int k = 0; k < intermediate; k++) {
	ocaLookUpBuffer::pointer feedbackLub;
	if (k == 0) {
	    feedbackLub
		= ocaCast<ocaLookUpBuffer>(ocaZeroBuffer::getSingleton());
	} else {
	    feedbackLub = feedback->getSharedLookUpBuffer();
	}
	target->makeCurrent();
	this->bind();

	cgGLSetTextureParameter(this->matrixACCUM, feedbackLub->getTextureId());
	cgGLEnableTextureParameter(this->matrixACCUM);

#define MATRIX_INDICES_R0C0 (matrixIndices)
#define MATRIX_INDICES_R0CN (matrixIndices + (DestRowIsY ? 1*4 : 3*4))
#define MATRIX_INDICES_RNC0 (matrixIndices + (DestRowIsY ? 3*4 : 1*4))
#define MATRIX_INDICES_RNCN (matrixIndices + 2*4)

#define MATRIX_INDEX_AR (ARowIsT ? 1 : 0)
#define MATRIX_INDEX_AC (ARowIsT ? 0 : 1)
#define MATRIX_INDEX_BR (BRowIsT ? 3 : 2)
#define MATRIX_INDEX_BC (BRowIsT ? 2 : 3)

      // Set up texture lookups.
	MATRIX_INDICES_R0C0[MATRIX_INDEX_AR] = (float)(0);
	MATRIX_INDICES_R0C0[MATRIX_INDEX_AC] = (float)(k + 0.5);
	MATRIX_INDICES_R0CN[MATRIX_INDEX_AR] = (float)(0);
	MATRIX_INDICES_R0CN[MATRIX_INDEX_AC] = (float)(k + 0.5);
	MATRIX_INDICES_RNC0[MATRIX_INDEX_AR] = (float)(rows);
	MATRIX_INDICES_RNC0[MATRIX_INDEX_AC] = (float)(k + 0.5);
	MATRIX_INDICES_RNCN[MATRIX_INDEX_AR] = (float)(rows);
	MATRIX_INDICES_RNCN[MATRIX_INDEX_AC] = (float)(k + 0.5);

	MATRIX_INDICES_R0C0[MATRIX_INDEX_BR] = (float)(k + 0.5);
	MATRIX_INDICES_R0C0[MATRIX_INDEX_BC] = (float)(0);
	MATRIX_INDICES_R0CN[MATRIX_INDEX_BR] = (float)(k + 0.5);
	MATRIX_INDICES_R0CN[MATRIX_INDEX_BC] = (float)(cols);
	MATRIX_INDICES_RNC0[MATRIX_INDEX_BR] = (float)(k + 0.5);
	MATRIX_INDICES_RNC0[MATRIX_INDEX_BC] = (float)(0);
	MATRIX_INDICES_RNCN[MATRIX_INDEX_BR] = (float)(k + 0.5);
	MATRIX_INDICES_RNCN[MATRIX_INDEX_BC] = (float)(cols);

      // Run vertex program.
	glDrawArrays(GL_QUADS, 0, 4);

	cgGLDisableTextureParameter(this->matrixACCUM);

      // Swap drawables
	if (k != 0) {
	    feedback->releaseLookUpBuffer();
	}
	ocaDrawableBuffer::pointer tmp = feedback;
	feedback = target;
	target = tmp;
    }

    if (!temporary.isNull()) {
	temporary->makeCurrent();
	cgGLDisableTextureParameter(this->matrixA);
	cgGLDisableTextureParameter(this->matrixB);
	cgGLDisableClientState(this->position);
	cgGLDisableClientState(this->index);
	this->unbind();
    }

    destination->makeCurrent();
    cgGLDisableTextureParameter(this->matrixA);
    cgGLDisableTextureParameter(this->matrixB);
    cgGLDisableClientState(this->position);
    cgGLDisableClientState(this->index);
    this->unbind();
}
