// -*- c++ -*-

// Tests the FFT in compressed format.

#include <ocaLookUpBuffer.h>
#include <ocaDrawableBuffer.h>
#include <ocaFactory.h>
#include <ocaError.h>
#include <ocaProgramFFT1c.h>
#include <ocaProgramFFT2c.h>
#include <ocaTimer.h>

#include <iostream>
#include <stdlib.h>
#include <time.h>
#include <math.h>

using std::cout;
using std::endl;

#define MAX_LOG_DIM	8
#define MIN_VALUE -2.0f
#define MAX_VALUE 5.0f

#define ABS(x)	((x) < 0 ? -(x) : (x))
#define EPSILON	0.0001
#define EQUAL(x1,x2)	(   (EPSILON > ABS(x1-x2))		\
			 || (ABS(EPSILON*x1) > ABS(x1-x2)))

#define TIMING_ONLY	0

/*
 * This method is Copyright (c) Paul Bourke 1993.
 * The original code can be retrieved from
 *	http://astronomy.swin.edu.au/~pbourke/analysis/dft/
 * This computes an in-place complex-to-complex FFT 
 * real and imag are the real and imaginary arrays of 2^m points.
 * dir =  1 gives forward transform
 * dir = -1 gives reverse transform 
*/
static short FFT(short int dir,long m,float *real,float *imag)
{
   long n,i,i1,j,k,i2,l,l1,l2;
   float c1,c2,treal,timag,t1,t2,u1,u2,z;

   /* Calculate the number of points */
   n = 1;
   for (i=0;i<m;i++) 
      n *= 2;

   /* Do the bit reversal */
   i2 = n >> 1;
   j = 0;
   for (i=0;i<n-1;i++) {
      if (i < j) {
         treal = real[i];
         timag = imag[i];
         real[i] = real[j];
         imag[i] = imag[j];
         real[j] = treal;
         imag[j] = timag;
      }
      k = i2;
      while (k <= j) {
         j -= k;
         k >>= 1;
      }
      j += k;
   }

   /* Compute the FFT */
   c1 = -1.0; 
   c2 = 0.0;
   l2 = 1;
   for (l=0;l<m;l++) {
      l1 = l2;
      l2 <<= 1;
      u1 = 1.0; 
      u2 = 0.0;
      for (j=0;j<l1;j++) {
         for (i=j;i<n;i+=l2) {
            i1 = i + l1;
            t1 = u1 * real[i1] - u2 * imag[i1];
            t2 = u1 * imag[i1] + u2 * real[i1];
            real[i1] = real[i] - t1; 
            imag[i1] = imag[i] - t2;
            real[i] += t1;
            imag[i] += t2;
         }
         z =  u1 * c1 - u2 * c2;
         u2 = u1 * c2 + u2 * c1;
         u1 = z;
      }
      c2 = (float)sqrt((1.0 - c1) / 2.0);
      if (dir == 1) 
         c2 = -c2;
      c1 = (float)sqrt((1.0 + c1) / 2.0);
   }

   /* Scaling for forward transform */
   if (dir == 1) {
      for (i=0;i<n;i++) {
         real[i] /= n;
         imag[i] /= n;
      }
   }
   
   return 1;
}


static void DoFFT(int lg_sizex, int lg_sizey)
{
    int i, k;
    int sizex, sizey;

    sizex = sizey = 1;
    for (i = 0; i < lg_sizex; i++) {
	sizex <<= 1;
    }
    for (i = 0; i < lg_sizey; i++) {
	sizey <<= 1;
    }

    cout << "Image size: " << sizex << "x" << sizey << endl;

    float *inbuf_original = new float[4*sizex*sizey];
    float *inbuf_real[4];
    float *inbuf_imag[4];
    for (i = 0; i < 4; i++) {
	inbuf_real[i] = new float[sizex*sizey];
	inbuf_imag[i] = new float[sizex*sizey];
    }
    float *outbuf_compressed = new float[4*sizex*sizey];
    float *outbuf_real = new float[4*sizex*sizey];
    float *outbuf_imag = new float[4*sizex*sizey];
    float *outbuf_compressed2 = new float[4*sizex*sizey];
    ocaTimer::pointer timer = ocaTimer::New();

    for (i = 0; i < 4*sizex*sizey; i++) {
	inbuf_original[i] = (  ((float)rand()/(float)RAND_MAX)
			     * (MAX_VALUE-MIN_VALUE) + MIN_VALUE);
    }
    for (i = 0; i < sizex*sizey; i++) {
	inbuf_real[0][i] = inbuf_original[4*i + 0];
	inbuf_real[1][i] = inbuf_original[4*i + 1];
	inbuf_real[2][i] = inbuf_original[4*i + 2];
	inbuf_real[3][i] = inbuf_original[4*i + 3];
	inbuf_imag[0][i] = 0;
	inbuf_imag[1][i] = 0;
	inbuf_imag[2][i] = 0;
	inbuf_imag[3][i] = 0;
    }

    ocaFactory::pointer factory = ocaFactory::getSingleton();

    timer->start();
    ocaLookUpBuffer::pointer original = factory->makeLookUpBuffer();
    original->setData(inbuf_original, sizex, sizey, 4);
    timer->stop();
    cout << "Time to load data: " << timer->getElapsedTime() << " sec" << endl;

    ocaDrawableBuffer::pointer frequencies_compressed
	= factory->makeDrawableBuffer(sizex, sizey, 4);

    cout << "Performing 1D FFT in S." << endl;
  // Compute FFT with Graphics card.
    timer->start();
    ocaProgramFFT1c::getSingleton()->execute(original, frequencies_compressed);
    glFinish();
    timer->stop();
    cout << "Time for GPU to perform FFT: " << timer->getElapsedTime()
	 << " sec" << endl;
    timer->start();
    frequencies_compressed->getData(outbuf_compressed);
    timer->stop();
    cout << "Time Retrieve FFT: " << timer->getElapsedTime() 
	 << " sec" << endl;

    for (i = 0; i < sizey; i++) {
	ocaProgramFFT1c::uncompress(sizex, 4,
				    (const float*)(outbuf_compressed+4*i*sizex),
				    outbuf_real + 4*i*sizex,
				    outbuf_imag + 4*i*sizex);
    }

  // Compute FFT (in place) with traditional function.
    timer->start();
    for (i = 0; i < sizey; i++) {
	FFT(1, lg_sizex, inbuf_real[0] + i*sizex, inbuf_imag[0] + i*sizex);
	FFT(1, lg_sizex, inbuf_real[1] + i*sizex, inbuf_imag[1] + i*sizex);
	FFT(1, lg_sizex, inbuf_real[2] + i*sizex, inbuf_imag[2] + i*sizex);
	FFT(1, lg_sizex, inbuf_real[3] + i*sizex, inbuf_imag[3] + i*sizex);
    }
    timer->stop();
    cout << "Time to do 1D FFT in software: " << timer->getElapsedTime() 
	 << " sec" << endl;

#if !TIMING_ONLY
    cout << "Checking data." << endl;
    for (k = 0; k < 4; k++) {
	for (i = 0; i < sizex*sizey; i++) {
	    if (   !EQUAL(inbuf_real[k][i]/sizex, outbuf_real[4*i+k]/sizex)
		|| !EQUAL(inbuf_imag[k][i]/sizex, outbuf_imag[4*i+k]/sizex) ) {
		cout << inbuf_real[k][i] << " + "
		     << inbuf_imag[k][i] << "i" << endl;
		cout << outbuf_real[4*i+k] << " + "
		     << outbuf_imag[4*i+k] << "i" <<endl;
		ocaRaiseError("FFTs don't agree!");
	    }
	}
    }
    cout << "FFT correct." << endl;

    cout << "Checking recompression." << endl;
    for (i = 0; i < sizey; i++) {
	ocaProgramFFT1c::compress(sizex, 4, outbuf_compressed2 + 4*i*sizex,
				  (const float *)(outbuf_real + 4*i*sizex),
				  (const float *)(outbuf_imag + 4*i*sizex));
    }
    for (i = 0; i < 4*sizex*sizey; i++) {
	if (outbuf_compressed[i] != outbuf_compressed2[i]) {
	    ocaRaiseError("Recompression incorrect.");
	}
    }
#endif //!TIMING_ONLY
    cout << "Recompression correct." << endl;

    cout << "Performing 2D FFT." << endl;
  // Compute FFT with Graphics card.
    timer->start();
    ocaProgramFFT2c::getSingleton()->execute(original, frequencies_compressed);
    glFinish();
    timer->stop();
    cout << "Time to compute 2D FFT on GPU: " << timer->getElapsedTime()
	 << " sec" << endl;
    timer->start();
    frequencies_compressed->getData(outbuf_compressed);
    timer->stop();
    cout << "Time to get data: " << timer->getElapsedTime() << " sec" << endl;

    ocaProgramFFT2c::uncompress(sizex, sizey, 4, 
				(const float *)outbuf_compressed,
				outbuf_real, outbuf_imag);

  // Compute FFT (in place) with traditional function.
    float *tmp_real = new float[sizey];
    float *tmp_imag = new float[sizey];
    timer->start();
    for (k = 0; k < 4; k++) {
	for (i = 0; i < sizex; i++) {
	    int j;
	    for (j = 0; j < sizey; j++) {
		tmp_real[j] = inbuf_real[k][j*sizex + i];
		tmp_imag[j] = inbuf_imag[k][j*sizex + i];
	    }
	    FFT(1, lg_sizey, tmp_real, tmp_imag);
	    for (j = 0; j < sizey; j++) {
		inbuf_real[k][j*sizex + i] = tmp_real[j];
		inbuf_imag[k][j*sizex + i] = tmp_imag[j];
	    }
	}
    }
    timer->stop();
    delete[] tmp_real;
    delete[] tmp_imag;

    cout << "Time to compute 2nd half of 2D FFT in software: "
	 << timer->getElapsedTime() << " sec" << endl;

#if !TIMING_ONLY
    cout << "Checking data." << endl;
    for (k = 0; k < 4; k++) {
	for (i = 0; i < sizex*sizey; i++) {
	    if (   !EQUAL(inbuf_real[k][i]/(sizex*sizey),
			  outbuf_real[4*i+k]/(sizex*sizey))
		|| !EQUAL(inbuf_imag[k][i]/(sizex*sizey),
			  outbuf_imag[4*i+k]/(sizex*sizey)) )
	    {
		cout << inbuf_real[k][i] << " + "
		     << inbuf_imag[k][i] << "i" << endl;
		cout << outbuf_real[4*i+k] << " + "
		     << outbuf_imag[4*i+k] << "i" << endl;
		ocaRaiseError("FFTs don't agree!");
	    }
	}
    }
    cout << "FFT correct." << endl;

    cout << "Checking recompression." << endl;
    ocaProgramFFT2c::compress(sizex, sizey, 4, outbuf_compressed2,
			      (const float *)outbuf_real,
			      (const float *)outbuf_imag);
    for (i = 0; i < 4*sizex*sizey; i++) {
	if (outbuf_compressed[i] != outbuf_compressed2[i]) {
	    cout << outbuf_compressed[i] << endl;
	    cout << outbuf_compressed2[i] << endl;
	    ocaRaiseError("Recompression incorrect.");
	}
    }
    cout << "Recompression correct." << endl;
#endif //!TIMING_ONLY

    delete[] inbuf_real[0];
    delete[] inbuf_real[1];
    delete[] inbuf_real[2];
    delete[] inbuf_real[3];
    delete[] inbuf_imag[0];
    delete[] inbuf_imag[1];
    delete[] inbuf_imag[2];
    delete[] inbuf_imag[3];
    delete[] outbuf_compressed;
    delete[] outbuf_real;
    delete[] outbuf_imag;
    delete[] outbuf_compressed2;
}


int CompressedFFT(int, char *[])
{
    srand((unsigned int)time(NULL));

    try {
	DoFFT(1, 1);
	DoFFT(rand()%MAX_LOG_DIM+1, rand()%MAX_LOG_DIM+1);
// // 	DoFFT(5, 5);
// 	DoFFT(9, 9);
    } catch (ocaError &error) {
	cout << "Got an error:" << endl
	     << error.getMessage() << endl;
	ocaObject::finalize();
	return 1;
    }
    ocaObject::finalize();

    return 0;
}
